def process_worldpop_data(input_rasters, output_file):
    """
    Merges and clips WorldPop rasters to a specified region using a shapefile.
    
    Parameters:
    - input_rasters: List of file paths to WorldPop raster files.
    - output_file: File path to save the processed raster.
    """
    import rasterio
    from rasterio.merge import merge
    #
    # Step 1: Merge raster files
    raster_list = [rasterio.open(raster) for raster in input_rasters]
    merged_raster, merged_transform = merge(raster_list)
    #
    # Step 2: Save the clipped raster to output_file
    output_meta = raster_list[0].meta.copy()
    output_meta.update({
        "driver": "GTiff",
        "height": merged_raster.shape[1],
        "width": merged_raster.shape[2],
        "transform": merged_transform
    })
    #
    with rasterio.open(output_file, "w", **output_meta) as dest:
        dest.write(merged_raster)
    #
    # Close all raster files
    for raster in raster_list:
        raster.close()
    #
    print(f"Processed raster saved to {output_file}")

def clip_raster(image_file, shape_file, output_file):
    import rasterio
    from rasterio.mask import mask
    import geopandas as gpd
    #
    img = rasterio.open(image_file)
    #
    region = gpd.read_file(shape_file)
    raster_crs = img.crs
    region = region.to_crs(raster_crs)
    #
    region_geom = [feature["geometry"] for feature in region.__geo_interface__["features"]]
    #
    clipped_raster, clipped_transform = mask(
        dataset=img,
        shapes=region_geom,
        crop=True
    )
    #
    output_meta = img.meta.copy()
    output_meta.update({
        "driver": "GTiff",
        "height": clipped_raster.shape[1],
        "width": clipped_raster.shape[2],
        "transform": clipped_transform
    })
    #
    with rasterio.open(output_file, "w", **output_meta) as dest:
        dest.write(clipped_raster)
    #
    img.close()

def url_worldpop_local():
    import re
    txts = [
        open("clim/WorldPop/WorldPop __ Population Density __ p1.html","r").read(),
        open("clim/WorldPop/WorldPop __ Population Density __ p2.html","r").read()
    ]
    url_pattern = r'"(https://hub.worldpop.org/geodata/summary\?id=.+?)"'
    return [elm for txt in txts for elm in re.findall(url_pattern, txt)]

def url_worldpop(website_url):
    import requests
    import re
    #
    try:
        # Make a GET request to fetch the website content
        response = requests.get(website_url)
        response.raise_for_status()  # Raise an error for bad HTTP responses
        #
        # Extract the HTML content
        html_content = response.text
        #
        # Regex to match URLs
        url_pattern = r'"(https://data.worldpop.org/GIS/Population_Density/Global_2000_2020_1km_UNadj/2020/.+/(.+_pd_2020_1km_UNadj.tif))"'
        #
        # Extract all URLs from the HTML content
        urls = re.findall(url_pattern, html_content)
        #
        if not len(urls) == 1:
            print("There is a problem with parsing",website_url)
            return
        url = urls[0]
        #
        try:
            # Make a GET request to fetch the file content
            response = requests.get(url[0])
            response.raise_for_status()  # Raise an error for bad HTTP responses
            #
            # Save the file content directly
            with open("clim/WorldPop/tif/%s" %(url[1]), "wb") as file:
                file.write(response.content)
            #
            print(f"File downloaded successfully and saved as '{url[1]}'.")
        #
        except requests.exceptions.RequestException as e:
            print(f"Error downloading file: {e}")
    #
    except requests.exceptions.RequestException as e:
        print(f"Error fetching {website_url}: {e}")

def tiff_to_ncdf():
    import numpy as np
    import xarray as xr
    import rioxarray
    #
    # Open the GeoTIFF file
    da = rioxarray.open_rasterio("clim/WorldPop/wpop_pd_2020_1km_UNadj_clipped.tif")
    #
    # Extract data and metadata
    data = np.ma.array(da[0, :, :], fill_value=da.rio.nodata, mask=da[0, :, :] == da.rio.nodata)
    lats = da.y.values
    lons = da.x.values
    #
    # Create a Dataset with CF-compliant metadata
    ds = xr.Dataset(
        data_vars={
            "dpop": xr.DataArray(
                data=data,
                dims=["lat", "lon"],
                coords={"lat": lats, "lon": lons},
                attrs={
                    "long_name": "Human Population Density",
                    "units": "people per km^2",
                    "coordinates": "lat lon",
                    "missing_value": da.rio.nodata
                    # "_FillValue": da.rio.nodata,
               },
            )
        },
        coords={
            "lat": ("lat", lats, {"units": "degrees_north", "standard_name": "latitude"}),
            "lon": ("lon", lons, {"units": "degrees_east", "standard_name": "longitude"}),
        },
        attrs={
            "description": "WorldPop Human Population Density in 2020 (per km^2)",
            "source": "clim/WorldPop/wpop_pd_2020_1km_UNadj_clipped.tif",
        },
    )
    #
    # Save the Dataset to a NetCDF file
    ds.to_netcdf(
        "clim/WorldPop/wpop_pd_2020_1km_UNadj_clipped.nc",
        engine="netcdf4",
        encoding={"dpop": {"zlib": True, "complevel": 9, "_FillValue": da.rio.nodata}},
    )
    #
    print("NetCDF file created successfully!")

# Usage
if __name__ == "__main__":
    if False:
        urls = url_worldpop_local()
        for url in urls:
            url_worldpop(url)
    #
    if True:
        import os
        input_rasters = ["clim/WorldPop/tif/%s" %(o) for o in os.listdir("clim/WorldPop/tif")]
        output_file = "clim/WorldPop/wpop_pd_2020_1km_UNadj.tif"
        process_worldpop_data(input_rasters, output_file)
    #
    # cdo griddes clim/CERRA/reanalysis-cerra-single-levels-2m_temperature-2015_clipped.nc > clim/griddes_CERRA_clipped.txt
    # cdo remapbil,clim/griddes_CERRA_clipped.txt clim/WorldPop/wpop_pd_2020_1km_UNadj_clipped.nc clim/WorldPop/wpop_pd_2020_1km_UNadj_clipped_remapbil.nc
